`timescale 1ns/1ps

module h_compact
#(
	parameter DATA_WIDTH = 32
)
(
    input      [DATA_WIDTH-1:0] data_i         ,
    input      [DATA_WIDTH-1:0] key_schedule0_i,
    input      [DATA_WIDTH-1:0] key_schedule1_i,
	
    output     [DATA_WIDTH-1:0] data_o
);

wire [(DATA_WIDTH/4)-1:0] data_upmost ;
wire [(DATA_WIDTH/4)-1:0] data_upper  ;  
wire [(DATA_WIDTH/4)-1:0] data_lower  ;  
wire [(DATA_WIDTH/4)-1:0] data_lowmost;

wire [(DATA_WIDTH/4)-1:0] key_schedule0_upmost ;
wire [(DATA_WIDTH/4)-1:0] key_schedule0_upper  ;
wire [(DATA_WIDTH/4)-1:0] key_schedule0_lower  ;
wire [(DATA_WIDTH/4)-1:0] key_schedule0_lowmost;

wire [(DATA_WIDTH/4)-1:0] key_schedule1_upmost ;
wire [(DATA_WIDTH/4)-1:0] key_schedule1_upper  ;
wire [(DATA_WIDTH/4)-1:0] key_schedule1_lower  ;
wire [(DATA_WIDTH/4)-1:0] key_schedule1_lowmost;

wire [(DATA_WIDTH/4)-1:0] s0_yield;
wire [(DATA_WIDTH/4)-1:0] s1_yield;
wire [(DATA_WIDTH/4)-1:0] s2_yield;
wire [(DATA_WIDTH/4)-1:0] s3_yield;

wire [(DATA_WIDTH/4)-1:0] z0;
wire [(DATA_WIDTH/4)-1:0] z1;
wire [(DATA_WIDTH/4)-1:0] z2;
wire [(DATA_WIDTH/4)-1:0] z3;

reg [(DATA_WIDTH/4)-1:0] mulBy5B_memory [0:(DATA_WIDTH*8)-1];
initial $readmemh("d:/Vivado/Projects/ip_repo/TwofishIP_1.0/src/mul_by_5B.txt", mulBy5B_memory);

reg [(DATA_WIDTH/4)-1:0] mulByEF_memory [0:(DATA_WIDTH*8)-1];
initial $readmemh("d:/Vivado/Projects/ip_repo/TwofishIP_1.0/src/mul_by_EF.txt", mulByEF_memory);

reg [(DATA_WIDTH/4)-1:0] permut_memory_q0 [0:(DATA_WIDTH*8)-1];
initial $readmemh("d:/Vivado/Projects/ip_repo/TwofishIP_1.0/src/q0_lookup.txt", permut_memory_q0);

reg [(DATA_WIDTH/4)-1:0] permut_memory_q1 [0:(DATA_WIDTH*8)-1];
initial $readmemh("d:/Vivado/Projects/ip_repo/TwofishIP_1.0/src/q1_lookup.txt", permut_memory_q1);

assign {data_lowmost,data_lower,data_upper,data_upmost} = data_i[(DATA_WIDTH-1):0];

assign {key_schedule0_lowmost,key_schedule0_lower,key_schedule0_upper,key_schedule0_upmost} = key_schedule0_i[(DATA_WIDTH-1):0];
assign {key_schedule1_lowmost,key_schedule1_lower,key_schedule1_upper,key_schedule1_upmost} = key_schedule1_i[(DATA_WIDTH-1):0];

assign s0_yield = permut_memory_q1[permut_memory_q0[permut_memory_q0[data_upmost]  ^ key_schedule1_upmost] ^ key_schedule0_upmost];
assign s1_yield = permut_memory_q0[permut_memory_q0[permut_memory_q1[data_upper]   ^ key_schedule1_upper] ^ key_schedule0_upper];
assign s2_yield = permut_memory_q1[permut_memory_q1[permut_memory_q0[data_lower]   ^ key_schedule1_lower] ^ key_schedule0_lower];
assign s3_yield = permut_memory_q0[permut_memory_q1[permut_memory_q1[data_lowmost] ^ key_schedule1_lowmost] ^ key_schedule0_lowmost];

assign z0 = s0_yield 				 ^ mulByEF_memory[s1_yield] ^ mulBy5B_memory[s2_yield] ^ mulBy5B_memory[s3_yield];
assign z1 = mulBy5B_memory[s0_yield] ^ mulByEF_memory[s1_yield] ^ mulByEF_memory[s2_yield] ^ s3_yield				 ;
assign z2 = mulByEF_memory[s0_yield] ^ mulBy5B_memory[s1_yield] ^ s2_yield				   ^ mulByEF_memory[s3_yield];
assign z3 = mulByEF_memory[s0_yield] ^ s1_yield					^ mulByEF_memory[s2_yield] ^ mulBy5B_memory[s3_yield];

assign data_o = {z3,z2,z1,z0};

endmodule