`timescale 1ns/1ps

module twofish_pipeline
#(
	parameter DATA_WIDTH = 128,
	parameter COUNT_TO   = 17 
)
(
	input 				         clk_i		 ,
	input 				         reset_ni	 ,
	input 				         enable_i	 ,
	
	input	[DATA_WIDTH/4-1:0]	 plaintext3_i,
	input	[DATA_WIDTH/4-1:0]	 plaintext2_i,
	input	[DATA_WIDTH/4-1:0]	 plaintext1_i,
	input	[DATA_WIDTH/4-1:0]	 plaintext0_i,
	
	input	[DATA_WIDTH/4-1:0]	 globalkey3_i,
	input	[DATA_WIDTH/4-1:0]	 globalkey2_i,
	input	[DATA_WIDTH/4-1:0]	 globalkey1_i,
	input	[DATA_WIDTH/4-1:0]	 globalkey0_i,
	
	output 						 	busy_o,	
	output	reg	[DATA_WIDTH/4-1:0]	cryptotext3_o,
	output	reg	[DATA_WIDTH/4-1:0]	cryptotext2_o,
	output	reg	[DATA_WIDTH/4-1:0]	cryptotext1_o,
	output	reg	[DATA_WIDTH/4-1:0]	cryptotext0_o
);

`include "clog_function.vh"
`include "plaintext_reorder.vh"
`include "cryptotext_reorder.vh"

wire [DATA_WIDTH-1:0] round_yeild ;

wire [DATA_WIDTH/4-1:0] plaintext_arranged3;
wire [DATA_WIDTH/4-1:0] plaintext_arranged2;
wire [DATA_WIDTH/4-1:0] plaintext_arranged1;
wire [DATA_WIDTH/4-1:0] plaintext_arranged0;

wire [(DATA_WIDTH/4)-1:0] 	R0_iw ;
wire [(DATA_WIDTH/4)-1:0] 	R1_iw ;
wire [(DATA_WIDTH/4)-1:0] 	R2_iw ;
wire [(DATA_WIDTH/4)-1:0] 	R3_iw ;

wire [(DATA_WIDTH/4)-1:0]   R0_ow ;
wire [(DATA_WIDTH/4)-1:0]   R1_ow ;
wire [(DATA_WIDTH/4)-1:0]   R2_ow ;
wire [(DATA_WIDTH/4)-1:0]   R3_ow ;

wire [(DATA_WIDTH/4*(COUNT_TO+3))-1:0]  K_key_odd   ; 
wire [(DATA_WIDTH/4*(COUNT_TO+3))-1:0]  K_key_even  ;

wire [(DATA_WIDTH/4)-1:0] K_inwhitening0 ;
wire [(DATA_WIDTH/4)-1:0] K_inwhitening1 ;
wire [(DATA_WIDTH/4)-1:0] K_inwhitening2 ;
wire [(DATA_WIDTH/4)-1:0] K_inwhitening3 ;

wire [(DATA_WIDTH/4)-1:0] K_outwhitening0;
wire [(DATA_WIDTH/4)-1:0] K_outwhitening1;
wire [(DATA_WIDTH/4)-1:0] K_outwhitening2;
wire [(DATA_WIDTH/4)-1:0] K_outwhitening3;

wire [clogb2(2*COUNT_TO+2):0] encrypt_counter;
wire encrypt_done;

assign plaintext_arranged3 = plaintext_reorder(plaintext0_i);
assign plaintext_arranged2 = plaintext_reorder(plaintext1_i);
assign plaintext_arranged1 = plaintext_reorder(plaintext2_i);
assign plaintext_arranged0 = plaintext_reorder(plaintext3_i);

assign busy_o = ~encrypt_done && enable_i;

assign encrypt_done = (encrypt_counter ===  2*COUNT_TO+3);

assign {K_inwhitening2,K_inwhitening0  } = K_key_even[(DATA_WIDTH/2)-1:0];
assign {K_inwhitening3,K_inwhitening1  } = K_key_odd [(DATA_WIDTH/2)-1:0];

assign {K_outwhitening2,K_outwhitening0} = K_key_even[DATA_WIDTH-1:DATA_WIDTH/2];
assign {K_outwhitening3,K_outwhitening1} = K_key_odd [DATA_WIDTH-1:DATA_WIDTH/2];

always @(posedge clk_i or negedge reset_ni) begin
	if(~reset_ni) begin
		cryptotext3_o <= 'b0;
    end
        cryptotext3_o <= cryptotext_reorder(R3_ow);
end

always @(posedge clk_i or negedge reset_ni) begin
	if(~reset_ni) begin
		cryptotext2_o <= 'b0;
    end
        cryptotext2_o <= cryptotext_reorder(R2_ow);

end

always @(posedge clk_i or negedge reset_ni) begin
	if(~reset_ni) begin
		cryptotext1_o <= 'b0;
    end
		cryptotext1_o <= cryptotext_reorder(R1_ow);


end

always @(posedge clk_i or negedge reset_ni) begin
	if(~reset_ni) begin
		cryptotext0_o <= 'b0;
    end
		cryptotext0_o <= cryptotext_reorder(R0_ow);

end

counter
#
(
    .COUNT_TO(2*COUNT_TO+4)
)
encrypt_counterINST
(
    .clk_i      (clk_i                    ),		
    .reset_ni	(reset_ni || encrypt_done ),
    .count_en_i (enable_i                 ),
    .count_o    (encrypt_counter          )
);

key_generator_compact
#
(
	.DATA_WIDTH(DATA_WIDTH/4),
	.COUNT_TO  (COUNT_TO + 3)
)
key_generator_compactINST
(
	.clk_i		   (clk_i		),	  
	.reset_ni	   (reset_ni	), 
    .keygen_en_i   (enable_i	),
    .K3_i		   (globalkey3_i),
	.K2_i          (globalkey2_i),
	.K1_i          (globalkey1_i),
	.K0_i          (globalkey0_i),
    .K_key_odd_o   (K_key_odd	),
    .K_key_even_o  (K_key_even	)
);

input_whitening
#(
    .DATA_WIDTH(DATA_WIDTH/4)
)
in_whiteningINST
(
	.enable_i (enable_i		  		),
	.P3_i	  (plaintext_arranged3  ),
	.P2_i     (plaintext_arranged2  ),
	.P1_i     (plaintext_arranged1  ),
	.P0_i	  (plaintext_arranged0	),
	.K0_i	  (K_inwhitening0 		),
	.K1_i	  (K_inwhitening1 		),
	.K2_i	  (K_inwhitening2 		),
	.K3_i	  (K_inwhitening3 		),
    .R0_o	  (R0_iw		  		),
    .R1_o	  (R1_iw		  		),
    .R2_o	  (R2_iw		  		),
    .R3_o	  (R3_iw		  		)
);

output_whitening
#(
    .DATA_WIDTH(DATA_WIDTH)
)
out_whiteningINST
(
	.clk_i		 (clk_i		       ),
	.reset_ni	 (reset_ni         ),
	.enable_i	 (enable_i		   ),
	.data_i      (round_yeild      ),
	.K0_i		 (K_outwhitening0  ),
	.K1_i		 (K_outwhitening1  ),
	.K2_i		 (K_outwhitening2  ),
	.K3_i		 (K_outwhitening3  ),
    .R0_o		 (R3_ow		       ),
    .R1_o		 (R2_ow		       ),
    .R2_o		 (R1_ow		       ),
    .R3_o		 (R0_ow		       )
);

genvar i;
generate
    wire [DATA_WIDTH/4-1:0] R0_aux[0:COUNT_TO-1];
    wire [DATA_WIDTH/4-1:0] R1_aux[0:COUNT_TO-1];
    wire [DATA_WIDTH/4-1:0] R2_aux[0:COUNT_TO-1];
    wire [DATA_WIDTH/4-1:0] R3_aux[0:COUNT_TO-1];
    
    wire [DATA_WIDTH/4-1:0] R0_auxo;
    wire [DATA_WIDTH/4-1:0] R1_auxo;
    wire [DATA_WIDTH/4-1:0] R2_auxo;
    wire [DATA_WIDTH/4-1:0] R3_auxo;
    
    assign R0_aux[0] = R0_iw;
    assign R0_auxo   = R0_aux[COUNT_TO-1];
    
    assign R1_aux[0] = R1_iw;
    assign R1_auxo   = R1_aux[COUNT_TO-1];
     
    assign R2_aux[0] = R2_iw;
    assign R2_auxo   = R2_aux[COUNT_TO-1];
    
    assign R3_aux[0] = R3_iw;
    assign R3_auxo   = R3_aux[COUNT_TO-1];
    
    for(i = 0; i < COUNT_TO-1 ; i = i+1) begin: rounds
        round_pipeline
        #(
            .DATA_WIDTH(DATA_WIDTH/4)
        )
        round_pipeline
        (	
			.clk_i		(clk_i	    ),
			.reset_ni	(reset_ni   ),
            .enable_i	(enable_i   ),		
            .R0_i		(R0_aux[i]	),		
            .R1_i		(R1_aux[i]  ),		
            .R2_i		(R2_aux[i]  ),		
            .R3_i		(R3_aux[i]  ),		
            .K_key_even	(K_key_even[(i+4)*(DATA_WIDTH/4)+:DATA_WIDTH/4]),
            .K_key_odd	(K_key_odd [(i+4)*(DATA_WIDTH/4)+:DATA_WIDTH/4]),

            .R0_o		(R0_aux[i+1] ),		
            .R1_o		(R1_aux[i+1] ),		
            .R2_o		(R2_aux[i+1] ),		
            .R3_o       (R3_aux[i+1] )
        );  
       
end

assign round_yeild = {R0_auxo,R1_auxo,R2_auxo,R3_auxo};  

endgenerate

	    
endmodule
	