`timescale 1ns/1ps

module key_generator_compact
#(
	parameter DATA_WIDTH   = 32,
	parameter COUNT_TO	   = 20
)
(
input 	  					      clk_i		    ,
input 						      reset_ni	    ,
input       					  keygen_en_i   ,
input      [(DATA_WIDTH*4)-1:0]   global_key_i  ,

output 	   [(COUNT_TO*DATA_WIDTH)-1:0] K_key_odd_o ,
output 	   [(COUNT_TO*DATA_WIDTH)-1:0] K_key_even_o	
);

`include "clog_function.vh"

wire [clogb2(COUNT_TO):0] key_index;
wire [(DATA_WIDTH/4)-1:0] key_index_mul2;
wire [(DATA_WIDTH/4)-1:0] key_index_mul2add1;
wire [DATA_WIDTH-1 : 0]   key_index_concat32bits_even;
wire [DATA_WIDTH-1 : 0]   key_index_concat32bits_odd;
reg  [DATA_WIDTH-1:0]  	  K_key_odd ; 
reg  [DATA_WIDTH-1:0]  	  K_key_even;

assign key_index_mul2 	  = 2*key_index;
assign key_index_mul2add1 = key_index_mul2 +1;
assign key_index_concat32bits_even = {4{key_index_mul2}};
assign key_index_concat32bits_odd  = {4{key_index_mul2add1}};

wire [DATA_WIDTH-1:0] globalkey_upper  ;
wire [DATA_WIDTH-1:0] globalkey_upmost ;
wire [DATA_WIDTH-1:0] globalkey_lower  ;
wire [DATA_WIDTH-1:0] globalkey_lowmost;

wire [DATA_WIDTH-1:0] h0_keygen_yield     ;
wire [DATA_WIDTH-1:0] h1_keygen_yield     ;
wire [DATA_WIDTH-1:0] h1_keygen_yield_rol8;
wire [DATA_WIDTH-1:0] pht1_yield_rol9	  ;
wire [DATA_WIDTH-1:0] pht_keygen0		  ;
wire [DATA_WIDTH-1:0] pht_keygen1		  ;

assign {globalkey_upmost,globalkey_upper,globalkey_lower,globalkey_lowmost} = (keygen_en_i) ? global_key_i[(DATA_WIDTH*4)-1:0] : 'b0;

assign h1_keygen_yield_rol8 = {h1_keygen_yield[23:0],h1_keygen_yield[(DATA_WIDTH-1):24]};
assign pht1_yield_rol9      = {pht_keygen1[22:0],pht_keygen1[(DATA_WIDTH-1):23]};

counter
#(
    .COUNT_TO(COUNT_TO+2)
)
counter_keyindex
(
    .clk_i      (clk_i                  ),
    .reset_ni	(reset_ni               ),
    .count_en_i (keygen_en_i      	    ),
    .count_o    (key_index              )
);

h_compact
#(
	.DATA_WIDTH(DATA_WIDTH)
)
h_compact0_keygenINST
(	
    .data_i			 (key_index_concat32bits_even),
	.key_schedule0_i (globalkey_upper	         ),
	.key_schedule1_i (globalkey_lowmost	         ), 
	                                             
    .data_o			 (h0_keygen_yield	         )
);

h_compact
#(
	.DATA_WIDTH(DATA_WIDTH)
)
h_compact1_keygenINST
(	
    .data_i			 (key_index_concat32bits_odd),
	.key_schedule0_i (globalkey_upmost	         ),
	.key_schedule1_i (globalkey_lower            ), 
	                                     
    .data_o			 (h1_keygen_yield	         )
);

pht 
#(
	.DATA_WIDTH(DATA_WIDTH)
)
pht_inst
(
    .T0_i	 (h0_keygen_yield		),
    .T1_i	 (h1_keygen_yield_rol8  ),
    .pht_o0  (pht_keygen0	  		),
    .pht_o1  (pht_keygen1	  		)
);

odd_key_memory
#(
	.DATA_WIDTH  (DATA_WIDTH  ),
	.DEPTH	     (COUNT_TO    )
)
odd_key_memoryINST
(
	.clk_i		(clk_i		),
    .reset_ni	(reset_ni	),
	.keygen_en_i(keygen_en_i),
	.key_index	(key_index	),
	.odd_key_i	(K_key_odd  ),
	.odd_key_o  (K_key_odd_o)
);

even_key_memory
#(
	.DATA_WIDTH	 (DATA_WIDTH  ),
	.DEPTH	   	 (COUNT_TO    )
)
even_key_memoryINST
(
	.clk_i		(clk_i		 ),
    .reset_ni	(reset_ni	 ),
	.keygen_en_i(keygen_en_i ),
	.key_index  (key_index	 ),
	.even_key_i (K_key_even  ),
	.even_key_o (K_key_even_o)
);

always @(posedge clk_i or negedge reset_ni) begin
	if(~reset_ni)		
		K_key_odd <= 'b0;
	else 
		if(keygen_en_i)
			K_key_odd <= pht1_yield_rol9;
end

always @(posedge clk_i or negedge reset_ni) begin
	if(~reset_ni)
		K_key_even <= 'b0;
	else
		if(keygen_en_i)
			K_key_even <= pht_keygen0;
end     
     
endmodule